/*	BioHaptix Arduino code for probe version 2.
	Written by Spencer Bondhus
        Copyright BioHaptix, 2013		*/

// TODO: double check use of analog pins and add delays to make sure timing doesn't interfere with analog reads
// TODO: add debug_all level code for calculating and displaying total run time for preconditioning and bioimpedance
// TODO: calibration routine

#include "definitions.h"
#include <MemoryFree.h>

/************************************************************************/
/**************************GLOBAL_VARIABLES******************************/
/************************************************************************/

debug_e debugLevel = DEBUG_MOST; // debug verbosity

String serialString = ""; // storage buffer for taking in serial commands
String command; // the command label being read in
long serialValues[MAX_SERIAL_NUMS]; // array for storing retrieved values
int numInd = 0; // index of received values

byte numElecPairs=0;
byte electrodePairs[MAX_ELECTRODE_PAIRS]; // array for storing received electrode pairs

int reservoirPressure = 55; // default reservoir pressure
int ddd=0;
int proportionairOffset = 0;
float scaleProp2Probe = 1; // calibrating prop write to probe sensor
float atmPsi = 14.5; // used to store atmospheric pressure
float probeatm=11.5;

unsigned long globalMillis = 0; // variable for keeping track of time
int loopDelay = 50; // delay between loops; set as a consequence of setVerbosity command
unsigned long pumpRun = 0;

boolean bElectrodePairsReceived;
boolean bPreConParamsReceived;
boolean bPreConCompleted;
boolean bTestRunning;
boolean bAborted;
boolean bAtmoCal;




pressureSensor_T pressureSensor_probe, pressureSensor_reservoir;

// preconditioning parameters
//    0: Pressure for preconditioning (kPa)
//    1: Number of preconditioning cycles (#)
//    2: Duration of precon pressurized (ms)
//    3: Duration of precon relaxed (ms)
preConParams_T preConParams;

bioRunParams_T bioRun;

/************************************************************************/
/**********************MAIN_PROGRAM_FUNCTIONS****************************/
/************************************************************************/

void setup() {
  if (debugLevel>DEBUG_SOME) Serial.println(F("DEBUG: Program started."));
  
  Serial.begin(SERIAL_BAUD);
  if (debugLevel>DEBUG_SOME) Serial.print(F("DEBUG: Serial started at ")); Serial.print(SERIAL_BAUD); Serial.println(F(" baud rate."));
  debugFreeMem();
  
  delay(500);
  initPins();
  initBools();
  initStructures();
 // initAtmosphere();
 // checkReservoir(reservoirPressure);
  
  if (debugLevel>DEBUG_NONE) Serial.println(F("DEBUG: Initializations complete."));
}

void loop(){
  delay(loopDelay);
   
  bAborted = false;
  
  runningChecks("loop");
  
 

  
  if (Serial.available()) {
    getSerial("loop");
    parseCommand();
    checkAndDelegate(command, numInd);
   
  }

}


/************************************************************************/
/********************BIOHAPTIX_MAIN_FUNCTIONS****************************/
/************************************************************************/

// will run through the entire preconditioning cylce, provided parameters have already been set
void runPreconditioning() {
  if (debugLevel>DEBUG_NONE) { Serial.println(F("DEBUG: subroutine startPreconditioning() started")); }
  debugFreeMem();
  
  digitalWrite(PIN_PROBE_LED_RED,LOW);
  if (!bPreConParamsReceived) { Serial.println(F("ERROR: preconditioning parameters must be uploaded first")); return; }
  if (preConParams.numCycles>0) {
    bTestRunning = true;
    
 //   adjustProp(preConParams.onPressure);//20
    
    unsigned long localMillis;
    for (int i=0;i<preConParams.numCycles;i++) {
      
      localMillis=millis();
      while((millis()-localMillis)<preConParams.onDuration){
        setProp(preConParams.onPressure);
        if (runningChecks("runPreconditioning")) return;
      }
      
      if (pressureFail()) return;
      
      localMillis=millis();
      while((millis()-localMillis)<preConParams.offDuration){
        setProp(0);
        if (runningChecks("runPreconditioning")) return;
      }
    } 
    
    Serial.println(F("preconditioningComplete"));
    resetAfterTest();
    bPreConCompleted = true;
  }
  else {
    Serial.println(F("preconditioningSkipped"));
  }
  return;
}

// will run through the entire bioimpedance cycle, periodically requesting new parameters from the PC
// will not run if electrode pairs have not been set
void runBioimpedance() {
  if (debugLevel>DEBUG_NONE) { Serial.println(F("DEBUG: subroutine startBioimpedance() started")); }
  debugFreeMem();
  
  digitalWrite(PIN_PROBE_LED_RED,LOW);
  if (!bElectrodePairsReceived) { Serial.println(F("ERROR: Electrode pairs must be uploaded first")); return; }
  
  bTestRunning = true;
  
 // adjustProp(preConParams.onPressure);

  Serial.println(F("requestBioParams"));
  unsigned long localMillis = millis();
  
  while (!bAborted) {
    if (Serial.available()) {
      getSerial("bioimpedeance");
      if (serialString.startsWith("runParams")) {
        localMillis = millis();
        parseCommand();
        checkAndDelegate(command, numInd);
      }
      else if (serialString.startsWith("runReady")) {
        localMillis = millis();
        Serial.println(F("runStarted"));
        for (byte i=0; i<numElecPairs; i++) {
          muxHold(electrodePairs[i],bioRun);
          if (bAborted) {bTestRunning=false; return;}
        }
        if (!bAborted) Serial.println(F("requestBioParams"));
      }
      else if (serialString.startsWith("bioimpedanceComplete")) {
        Serial.println(F("testingComplete"));
        bTestRunning=false;
        break;
      }
      else if (serialString.startsWith("ABORT")) {
        Serial.println(F("ABORTED: runBioimpedance"));
        bAborted=true;
        return;
      }
      else {
        Serial.println(F("RETRY: unrecognized command; try again. [runBioimpedance]"));
      }
    } // end if serial

    if ((millis()-localMillis)>TIMEOUT_MS) {
      Serial.println(F("ERROR: timeout in bioimpedanceComplete."));
      bTestRunning=false;
      return;
    }
  } // end while

  resetAfterTest();
  return;
}

// runs both preconditioning and bioimpedance. can be run from serial command or pressing the switch on the probe
void runFullCycle() {
  if (debugLevel>DEBUG_NONE) { Serial.println(F("DEBUG: subroutine startFullCycle() started")); }
  debugFreeMem();
  
  digitalWrite(PIN_PROBE_LED_RED,LOW);
  if (!bElectrodePairsReceived) { Serial.println(F("ERROR: electrode pairs must be uploaded first")); return; }
  if (!bPreConParamsReceived) { Serial.println(F("ERROR: preconditioning parameters must be uploaded first")); return; }

  bTestRunning = true;
  
  if (!bAborted) runPreconditioning();
  //if (!bAborted) runBioimpedance();

  //resetAfterTest();
  return;
}

/************************************************************************/
/*******************BIOHAPTIX_HELPER_FUNCTIONS***************************/
/************************************************************************/

// used in various positions in the program. Used to
//  (1) make LEDs flash on set period,
//  (2) check and refill the vacuum reservoir, and
//  (3) check if the ABORT command has been issued, emanding an end to the current process
boolean runningChecks(String functionName) {
  if (debugLevel==DEBUG_ALL) { Serial.println(F("DEBUG: subroutine runningChecks() started")); }
  setLeds();
 
//!!!!!!!!!!!!!!!! Hey, I make revisons here, remember to put the sentence below back!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  checkReservoir(reservoirPressure);
  if (bTestRunning) { // only check for abort command if a test is running
    if (checkAbort(functionName)) {
      return true;
    }
  }
  else checkSwitch(); // only check for switch presses if tests are not running
  
  return false;
}
  
// monitors serial command
boolean checkAbort(String functionName) {
  if (debugLevel==DEBUG_ALL) { Serial.println(F("DEBUG: subroutine checkAbort() started")); }
  if (Serial.available()) {
    getSerial("checkabort");
    if (serialString.startsWith("ABORT")) {
      
      Serial.print(F("ABORTED: "));
      Serial.println(functionName);
      
      bAborted = true;
      bTestRunning = false;
      return true;
    }
  }
  return false;
}

void setLeds() {
  if (debugLevel==DEBUG_ALL) { Serial.println(F("DEBUG: subroutine muxHold() started")); }
  
  if (bTestRunning) {
    if ((millis()-globalMillis)>(500/BLINK_RATE_TESTING)) {
      digitalWrite(PIN_PROBE_LED_GREEN,1-digitalRead(PIN_PROBE_LED_GREEN));
      globalMillis=millis();
    }
  }
  else if (bElectrodePairsReceived && bPreConParamsReceived) { digitalWrite(PIN_PROBE_LED_GREEN,HIGH); }
  else { digitalWrite(PIN_PROBE_LED_GREEN,LOW); }
}

void checkReservoir(int desiredPressure) {
  if (debugLevel==DEBUG_ALL) { Serial.println(F("DEBUG: subroutine checkReservoir() started")); }
  
  int currentPressure = honeywellPressure(pressureSensor_reservoir);
  
  if (millis()-pumpRun>MIN_PUMP_RUNTIME) {
    if (currentPressure>desiredPressure+PRESSURE_TOLERANCE) { digitalWrite(PIN_PUMP_POWER,LOW); }
    else if (currentPressure<desiredPressure-PRESSURE_TOLERANCE) { digitalWrite(PIN_PUMP_POWER,HIGH); pumpRun=millis(); }
  }
}

void checkSwitch() {
  if (debugLevel==DEBUG_ALL) { Serial.println(F("DEBUG: subroutine checkSwitch() started")); }
  
  // flash LED at 5Hz for 5 seconds; if user presses switch again in that time, runFullCycle()
  if (!digitalRead(PIN_PROBE_SWITCH)) {
    for (int i=0; 1000*i<(BLINK_TIME_SWITCH*BLINK_RATE_SWITCH*2); i++) {
      digitalWrite(PIN_PROBE_LED_GREEN,1-digitalRead(PIN_PROBE_LED_GREEN));
      if (digitalRead(PIN_PROBE_SWITCH)==HIGH) { return; }
      delay(500/BLINK_RATE_SWITCH);
    }
    runFullCycle();
  }
}

void resetAfterTest() {
  bTestRunning = false;
  digitalWrite(PIN_MUX_ENABLE,LOW); // MUX should be disabled
  setProp(0);
}



void setPropHectoPascals(int desiredPressure) {
 
  if (debugLevel>DEBUG_NONE) { Serial.println(F("DEBUG: subroutine setProp() started")); }
  debugFreeMem();
  Serial.print(F("desiredproppressure:"));
  Serial.println(desiredPressure);
  int propDelay=100, solDelay=10; // these delays are used to make sure prop is operating as efficiently as possible
  
  //desiredPressure *= scaleProp2Probe; // adjust prop based on calibration from adjustProp()

    Serial.print(F("desiredprobepressure:"));
  Serial.println(desiredPressure);
  int currentPressure = honeywellPressureforprobe(pressureSensor_probe);

  Serial.print(F("currentprobepressure:"));
  Serial.println(currentPressure);
  
  
//  if (currentPressure < (desiredPressure/10-PRESSURE_TOLERANCE) ) { propDelay=100; solDelay=10; }//add "/10" here
   //else if (currentPressure > (desiredPressure/10+PRESSURE_TOLERANCE) ) { propDelay=10; solDelay=500; }//add "/10" here
   

int desiredPressure_revised1=desiredPressure;
 int prop255 = (int) (((5.076*desiredPressure_revised1)/10.0)+0.203);
 if (prop255>255) prop255=255;
 if (prop255<0) prop255=0;
 analogWrite(PIN_PROP_WRITE,prop255); delay(propDelay);
 Serial.print(F("propsetpressure:"));
 Serial.println(prop255);

  //solenoid toggled to prevent valve from excessive fluttering
  digitalWrite(PIN_SOLENOID_POWER,HIGH); delay(solDelay); digitalWrite(PIN_SOLENOID_POWER,LOW);

}


//setProp will set the PROPORTION-AIR proportionary valve to desiredPressure, given in kiloPascals
void setProp(int desiredPressure) {
  setPropHectoPascals(desiredPressure*10);
  
}
  
boolean pressureFail() {
  
  for (int i=0; i<3; i++) {
    int propPressure = proportionairPressure(PIN_PROP_READ);
    int probePressure = honeywellPressureforprobe(pressureSensor_probe);
    if ((propPressure>1) && (probePressure>1)) return false;
  }
  
  digitalWrite(PIN_PROBE_LED_RED,HIGH);
  bTestRunning=false;
  bAborted=true;
  setProp(0);
  Serial.println(F("ERROR: No pressure observed at probe tip. Run Aborted."));
  return true;
}

/*void adjustProp(int desiredPressure) {
  setProp(desiredPressure);
  delay(200);
  
  int probePressure = honeywellPressureforprobe(pressureSensor_probe);
Serial.print(F("The adjust pressure from the probe is ["));
  Serial.print(probePressure);
  Serial.println(F("]"));
  int propPressure = proportionairPressure(PIN_PROP_READ);
  Serial.print(F("The adjust pressure from the proportionair is ["));
  Serial.print(propPressure);
  Serial.println(F("]"));
  float pressureRatio = ( ( (float) probePressure / (float) propPressure ) ) ;
  
  if ( (pressureRatio>0.5) && (pressureRatio<1.5) ) {
    scaleProp2Probe = pressureRatio;
  }
  else {
    Serial.println(F("DEBUG: proportionair not within 50% of probe; using default"));
    scaleProp2Probe = 1.0;
  }
     Serial.print(F("The prop to probe scale is:"));
     Serial.println(scaleProp2Probe);
  setProp(0);
  delay(50);
  
  return;
}*/

void muxHoldAndCycleHectoPascals(long propDelayms, long holdus, int pressurehpa) {
  if (propDelayms!=0)
  {setPropHectoPascals(pressurehpa);
  delay(propDelayms);}
  for(byte i = 0; i < numElecPairs; i++) {
    switchElec(electrodePairs[i]);
    unsigned long localMicros = micros();
    digitalWrite(PIN_MUX_ENABLE, HIGH);
    while((micros() - localMicros) < holdus) {
      if(runningChecks("muxHold")) return;
    }
 // Serial.println(F("Cycle Pin here is high"));
    digitalWrite(PIN_MUX_ENABLE, LOW);
    delay(220); // have to throttle for the PicoScope to pick it up
  }
  Serial.println(F("muxHoldAndCycleComplete"));
}

void muxHoldHectoPascals(byte elecPair, long propDelayms, long holdus, int pressurehpa) {
  
  if (propDelayms!=0)
  {setPropHectoPascals(pressurehpa) ;}
   delay(propDelayms);

  switchElec(elecPair);
  

  /*
  //////////////////
     unsigned long  timetracker3 = micros();
    //  df= (timetracker3- timetracker2)/1000000;
   Serial.print(F("Time after I set the Proportionair pressure is:"));
  Serial.println(timetracker3);
  ////////////////////////////////////
 
  int laterprobepressure=honeywellPressureforprobe(pressureSensor_probe);
    Serial.print(F("laterprobepressure:"));
  Serial.println(laterprobepressure);
   
    float laterproppressure=proportionairPressure(PIN_PROP_READ);
    Serial.print(F("laterproppressure:"));
  Serial.println(laterproppressure);
  //////////////////
     unsigned long timetracker = micros();
    Serial.print(F("The time before the Pin write also before holding time :"));
  Serial.println(timetracker);
  ///////////////////////////////
   */
  unsigned long localMicros = micros();
  Serial.println(F("Pin here is high"));
   digitalWrite(PIN_MUX_ENABLE, HIGH);
  while((micros() - localMicros) < holdus) {
    //Probably need to optimize the check routine to deal with mux hold times that are shorter
    //than the current runningChecks() run time once we get above 100khz or optimize period
    //counts to some low number (e.g., 100 periods @ 1Mhz = 0.1ms hold time.  If we optimize
    //to below 100 periods then the hold time is likely to be significantly less than the
    //run time of the current runningCheck function
    
    
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!I am testing here, remember to add back to code again!!!!!!!!!!!!!!!!!!!!
    if (runningChecks("muxHold")) return;
  }
  /*
  //////////////////
     timetracker = micros();
    Serial.print(F("The time after recording the data and set pin low again :"));
  Serial.println(timetracker);
  ///////////////////////////////
 */
  digitalWrite(PIN_MUX_ENABLE, LOW);
  delay(120);
  Serial.println(F("muxHoldComplete"));
}

void muxHold(byte elecPair, bioRunParams_T thisRun) {
  if (debugLevel>DEBUG_NONE) { Serial.println(F("DEBUG: subroutine muxHold() started")); }
  muxHoldHectoPascals(elecPair, 0, thisRun.holdus, thisRun.probePressure*10);
  /*
  setProp(thisRun.probePressure); // adjust prop to correct value
  
  unsigned long localMicros = micros();
  switchElec(elecPair);
  digitalWrite(PIN_MUX_ENABLE,HIGH);
  while ((micros()-localMicros)<thisRun.holdus) {
    if (runningChecks("muxHold")) return;
  }
  digitalWrite(PIN_MUX_ENABLE,LOW);
  */
  pressureFail(); // check if the prop was below pressure; if so, signal aborted test.
}

void switchElec(byte elecPair) {
  if (debugLevel>DEBUG_NONE) { Serial.println(F("DEBUG: subroutine switchElec() started")); }
  debugFreeMem();
  
  byte binMask=1;
  if (debugLevel>DEBUG_SOME) {
    Serial.print(F("DEBUG: Mux1 (dec): ")); Serial.print(elecPair & 15);
    Serial.print(F("; Mux1 (bin): "));
    Serial.print(  (elecPair & binMask) > 0); Serial.print(F(", ")); binMask *= 2;
    Serial.print(  (elecPair & binMask) > 0); Serial.print(F(", ")); binMask *= 2;
    Serial.print(  (elecPair & binMask) > 0); Serial.print(F(", ")); binMask *= 2;
    Serial.println((elecPair & binMask) > 0); binMask *= 2;
    
    Serial.print(F("DEBUG: Mux2 (dec): ")); Serial.print((elecPair>>4) & 15);
    Serial.print(F("; Mux2 (bin): "));
    Serial.print(  (elecPair & binMask) > 0); Serial.print(F(", ")); binMask *= 2;
    Serial.print(  (elecPair & binMask) > 0); Serial.print(F(", ")); binMask *= 2;
    Serial.print(  (elecPair & binMask) > 0); Serial.print(F(", ")); binMask *= 2;
    Serial.println((elecPair & binMask) > 0); binMask = 1;
  }
  
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX1+0,HIGH); } else { digitalWrite(PINOFFSET_MUX1+0,LOW); }; binMask *= 2;
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX1+1,HIGH); } else { digitalWrite(PINOFFSET_MUX1+1,LOW); }; binMask *= 2;
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX1+2,HIGH); } else { digitalWrite(PINOFFSET_MUX1+2,LOW); }; binMask *= 2;
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX1+3,HIGH); } else { digitalWrite(PINOFFSET_MUX1+3,LOW); }; binMask *= 2;
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX2+0,HIGH); } else { digitalWrite(PINOFFSET_MUX2+0,LOW); }; binMask *= 2;
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX2+1,HIGH); } else { digitalWrite(PINOFFSET_MUX2+1,LOW); }; binMask *= 2; 
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX2+2,HIGH); } else { digitalWrite(PINOFFSET_MUX2+2,LOW); }; binMask *= 2; 
  if (elecPair & binMask) { digitalWrite(PINOFFSET_MUX2+3,HIGH); } else { digitalWrite(PINOFFSET_MUX2+3,LOW); };
}

/************************************************************************/
/**************************SERIAL_FUNCTIONS******************************/
/************************************************************************/

void getSerial(String calledFrom)
{
  if (debugLevel>DEBUG_SOME) {
    Serial.print(F("DEBUG: subroutine getSerial() called from: "));
    Serial.println(calledFrom);
  }
  debugFreeMem();
  char inByte = ' ';
  serialString = "";
  numInd=0;
  while (Serial.available() > 0)
  {
    inByte = Serial.read(); delay(2);
    serialString.concat(inByte);
  }
  if (debugLevel>DEBUG_SOME) { Serial.print(F("DEBUG: Serial string received: ")); Serial.println(serialString); }
 
}

void parseCommand()
{
  
  if (debugLevel>DEBUG_SOME) Serial.println(F("DEBUG: subroutine parseCommand() started"));
  debugFreeMem();
  
  unsigned int nextDelim = 0; // index in remainingString of next delimiter
  String remainingString = serialString; // command string that is unread so far
  serialString = ""; // reset serial string
  
  // parse out command label
  nextDelim = remainingString.indexOf(DELIM_COM);
  command = remainingString.substring(0,nextDelim);
  if (debugLevel>DEBUG_SOME)
  {
    Serial.print(F("DEBUG: Serial command received: "));
    Serial.println(command);
  }

  // parse out values
  if (remainingString.length()>nextDelim)
  {
    do
    {
      remainingString = remainingString.substring(nextDelim+1,remainingString.length());
      nextDelim = remainingString.indexOf(DELIM_NUM);
      serialValues[numInd] = str2long(remainingString.substring(0,nextDelim));
      numInd++;
    } while (nextDelim<remainingString.length());
  }

  if (debugLevel>DEBUG_SOME && numInd>0)
  {
    Serial.print(F("DEBUG: Serial values received: "));
    Serial.print(serialValues[0]);
    if (numInd>1)
    {
      for (int i=1; i<numInd; i++)
      {
        Serial.print(F(", "));
        Serial.print(serialValues[i]);
      }
    }
    Serial.println();
  }
}

boolean checkAndDelegate(String command, int numParams)
{
  if (debugLevel>DEBUG_SOME) Serial.println(F("DEBUG: subroutine checkAndDelegate() started"));
  debugFreeMem();

  boolean retVal = true;
  if (command=="setPrecondition") {
    if (checkNumParams(command,numParams,NUMPARAMS_SETPRECON)) {
      if (       checkRange(command, MIN_PRESSURE,        reservoirPressure-MAX_PRESSURE_OFFSET, serialValues[0])) { // MIN_PRESSURE to 60 kPa
        if (     checkRange(command, 0,                   MAX_PRECON_CYCLES,                     serialValues[1])) { // 0 to 99 cycles (0 means skip precon)
          if (   checkRange(command, MIN_PRECON_DURATION, MAX_PRECON_DURATION,                   serialValues[2])) { // 0 to 1 minutes (0 to 60000ms)
            if ( checkRange(command, MIN_PRECON_DURATION, MAX_PRECON_DURATION,                   serialValues[3])) { // 0 to 1 minutes (0 to 60000ms)
              preConParams.onPressure  = (byte) serialValues[0];
              preConParams.numCycles   = (byte) serialValues[1];
              preConParams.onDuration  = (unsigned int) serialValues[2];
              preConParams.offDuration = (unsigned int) serialValues[3];
              Serial.println(F("preconParamsReceived"));
              bPreConParamsReceived = true;
            }
          }
        }
      }
    }
  }
  else if (command=="setVerbosity") {
    if (checkNumParams(command,numParams,NUMPARAMS_SETVERB)) {
      if (checkRange(command, 0, 3, serialValues[0])) {
        debugLevel = (debug_e) serialValues[0];
        if       (debugLevel==DEBUG_ALL)  loopDelay = 2000;
        else if  (debugLevel==DEBUG_MOST) loopDelay = 1000;
        else if  (debugLevel==DEBUG_SOME) loopDelay = 100;
        else                              loopDelay = 20;
        if (debugLevel>DEBUG_SOME) { Serial.print(F("DEBUG: Debug set to value: ")); Serial.println(debugLevel); }
      }
    }
  }
  else if (command=="runPreconditioning") {
    if (checkNumParams(command,numParams,NUMPARAMS_RUNPRECON)) {
      runPreconditioning();
    }
  }
  else if (command=="runBioimpedance") {
    if (checkNumParams(command,numParams,NUMPARAMS_RUNBIOIMP)) {
      runBioimpedance();
    }
  }
  else if (command=="runParams") {
    if (checkNumParams(command,numParams,NUMPARAMS_RUNPARAMS)) {
      if (  checkRange( command, MIN_RUNTIME,  MAX_LONG,                              serialValues[0])) {
        if (checkRange( command, MIN_PRESSURE, reservoirPressure-MAX_PRESSURE_OFFSET, serialValues[1])) {
          bioRun.holdus = serialValues[0];
          bioRun.probePressure = serialValues[1];
          Serial.println(F("paramsReceived"));
        }
      }
    }
  }
  else if (command=="runFullCycle") {
    if (checkNumParams(command,numParams,NUMPARAMS_RUNFULL)) {
      runFullCycle();
    }
  }
  else if (command=="muxEnable") {
    if (checkNumParams(command,numParams,0)) {
      digitalWrite(PIN_MUX_ENABLE,HIGH);
    }
  }
  else if (command=="muxDisable") {
    if (checkNumParams(command,numParams,0)) {
      digitalWrite(PIN_MUX_ENABLE,LOW);
    }
  }
  else if (command=="switchElec") {
    if (checkNumParams(command,numParams,NUMPARAMS_SWITCHELEC)) {
      if (  checkRange( command, 0, 15, serialValues[0])) {
        if (checkRange( command, 0, 15, serialValues[1])) {
          switchElec( (byte) ( ( serialValues[0] & 15) | ( ( serialValues[1] & 15 ) << 4 ) ) );
        }
      }
    }
  }
  else if (command == "muxHoldAndCycleHectoPascals") { // muxHoldAndCycleHectoPascals/<time to hold us>,<pressure hPa>,<prop delay ms>
    muxHoldAndCycleHectoPascals(serialValues[2], serialValues[0], serialValues[1]);
  }
  else if (command == "muxHoldHectoPascals") { // muxHoldHectoPascals/<time to hold us>,<pressure hPa>,<prop delay ms>,<electrode 1>,<electrode 2>
    byte electrodePair = (byte) ( ( serialValues[3] & 15 ) | ( ( serialValues[4] & 15 ) << 4 ) );
    muxHoldHectoPascals(electrodePair, serialValues[2], serialValues[0], serialValues[1]);
  }
  else if (command=="muxHold") {
    if (checkNumParams(command,numParams,NUMPARAMS_SETPRECON)) {
      if (       checkRange(command, 0, MAX_LONG, serialValues[0])) { // time to hold (us)
        if (     checkRange(command, 1, 40,       serialValues[1])) { // pressure to hold at (kPa)
          if (   checkRange(command, 0, 15,       serialValues[2])) { // electrode 1 (0-15)
            if ( checkRange(command, 0, 15,       serialValues[3])) { // electrode 2 (0-15)
              bioRunParams_T runThis;
              runThis.holdus = serialValues[0];
              runThis.probePressure = serialValues[1];
              
              byte electrodePair=(byte) ( ( serialValues[2] & 15) | ( ( serialValues[3] & 15 ) << 4 ) );
              muxHold(electrodePair,runThis);
            }
          }
        }
      }
    }
  }
  else if (command=="setElectrodes") {
    if (checkNumParams(command,numParams,NUMPARAMS_SETELEC)) {
      if (checkRange(command,0,MAX_ELECTRODE_PAIRS,serialValues[0])) {
        receiveElectrodePairs(serialValues[0]);
      }
    }
  }
  else if (command=="electrodeSet") {
    if (checkNumParams(command,numParams,NUMPARAMS_ELECSET)) {
      if (    checkRange( command, 0, MAX_ELECTRODE_PAIRS, serialValues[0])) {
        if (  checkRange( command, 0, 15,                  serialValues[1])) {
          if (checkRange( command, 0, 15,                  serialValues[2])) {
            electrodePairs[serialValues[0]]=(byte) ( ( serialValues[1] & 15) | ( ( serialValues[2] & 15 ) << 4 ) );
            Serial.println(F("setReceived"));
            retVal=true;
          }
        }
      }
    }
  }
  else if (command=="setReservoir") {
    if (checkNumParams(command,numParams,NUMPARAMS_SETRESERVOIR)) {
      if (checkRange(command,MIN_RESERVOIR,MAX_RESERVOIR,serialValues[0])) {
        reservoirPressure = serialValues[0];
      }
    }
  }
  else if (command == "setPropHectoPascals") {
    if (checkNumParams(command, numParams, NUMPARAMS_PROPORTIONAIR)) {
      if (checkRange(command, MIN_PROPORTIONAIR, MAX_PROPORTIONAIR, serialValues[0])) {
	setPropHectoPascals(serialValues[0]);
	Serial.println(F("setPropComplete"));
      }
    }
  }
  else if (command=="electrodePairsComplete") {
    if (checkNumParams(command,numParams,NUMPARAMS_ELECCOMPLETE)) {
      //response handled in receiveElectrodePairs()
    }
  }
  else if (command=="checkAtmoCal") {
    if (bAtmoCal) Serial.println(F("TRUE"));
    else Serial.println(F("FALSE"));
  }
  else if (command=="PCReady") {
    Serial.println(F("ArduinoReady"));
  }
  else if (command=="ABORT") {
    Serial.println(F("ABORTED: mainLoop"));
  }
  else if (command=="STOP") {
 Serial.println(F("stop"));
 delay(200);
  }
  else if (command=="RESET") {
    Serial.println(F("remove the probe tip from the tissue"));
    delay(200);
    initPins();
    initBools();
 //   initAtmosphere();
  }
  else
  {
    Serial.println(F("ERROR: command not recognized"));
  }
  return retVal;
}

int receiveElectrodePairs(int numPairs) {
  if (debugLevel>DEBUG_SOME) Serial.println(F("DEBUG: subroutine receiveElectrodePairs() started"));
  debugFreeMem();
  
  numElecPairs=(byte) numPairs;
  int pairNum=0;
  unsigned long localMillis;
  localMillis = millis();
  
  Serial.println(F("startSending"));
  while (pairNum<=numPairs) {
    if (Serial.available()) {
      getSerial("receiveelectrodepairs");
      if (serialString.startsWith("electrodeSet")) {
        localMillis=millis();
        parseCommand();
        if (checkAndDelegate(command, numInd)) {
          if (pairNum==serialValues[0]) pairNum++;
          else {
            Serial.print(F("ERROR: incorrect pair num; expected "));
            Serial.println(pairNum);
          }
        }
      }
      else if (serialString.startsWith("electrodePairsComplete")) {
        if (pairNum==numPairs) {
          Serial.println(F("electrodePairsReceived"));
          bElectrodePairsReceived=true;
          return 0;
        }
        else {
          Serial.println(F("ERROR: received less electrode pairs than expected."));
          return 0;
        }
      }
      else if (serialString.startsWith("ABORT")) {
        Serial.println(F("ABORTED: receiveElectrodePairs"));
        resetAfterTest();
        return 0;
      }
      else {
        Serial.println(F("RETRY: electrodePairsFailure"));
        serialString = "";
        return 0;
      }
    }
    
    if ((millis()-localMillis)>TIMEOUT_MS) {
      Serial.println(F("ERROR: timeout in receiveElectrodePairs."));
      return 0;
    }
  }
  Serial.println(F("ERROR: exceeded expected number of electrode pairs."));
}

/************************************************************************/
/**********************SERIAL_HELPER_FUNCTIONS***************************/
/************************************************************************/

// checks the number of parameters received to make sure it's okay
boolean checkNumParams(String cmd, int numDel, int numExp)
{
  if (numExp!=numDel) {
    Serial.print(F("ERROR: incorrect number of arguments received ("));
    Serial.print(numDel);
    Serial.print(F("/"));
    Serial.print(numExp);
    Serial.print(F(") ["));
    Serial.print(cmd);
    Serial.println(F("]"));
    return false;
  }
  else return true;
}

boolean checkRange(String cmd, long expLo, long expHi, long delVal) 
{
  if (delVal<expLo || delVal>expHi) {
    Serial.print(F("ERROR: parameter out of range ("));
    Serial.print(delVal);
    Serial.print(F("/"));
    Serial.print(expLo);
    Serial.print(F("-"));
    Serial.print(expHi);
    Serial.print(F(") ["));
    Serial.print(cmd);
    Serial.println(F("]"));
    return false;
  }
  else return true;
}

/************************************************************************/
/**************************HELPER_FUNCTIONS******************************/
/************************************************************************/

// str2long: convert string to long integer
long str2long(String str) {
  char carray[20];
  str.toCharArray(carray,20);
  return atol(carray);
} // END str2long()

// Pport = (Vout - 0.1*Vsupply)*(Pmax-Pmin)/(0.8*Vsupply)+Pmin
int honeywellPressure(pressureSensor_T sensorStruct) {
  return honeywellPressureHectoPascals(sensorStruct) / 10;
} // END honeywellPressure()

int honeywellPressureforprobe(pressureSensor_T sensorStruct) {
  return honeywellPressureHectoPascalsforprobe(sensorStruct) / 10;
} // END honeywellPressure()


int honeywellPressureHectoPascals(pressureSensor_T sensorStruct) {
  int pinValue = analogRead(sensorStruct.Pin);
  float pinVolts = 5.0*pinValue/1023.0;
  
  float Pport_abs = (pinVolts - 0.1*sensorStruct.Vsupply) * (sensorStruct.Pmax - sensorStruct.Pmin) / (0.8*sensorStruct.Vsupply) + sensorStruct.Pmin;
  float Pport_gage = psi2kPa(atmPsi-Pport_abs);
  return (int) abs(Pport_gage)*10;
}

int honeywellPressureHectoPascalsforprobe(pressureSensor_T sensorStruct) {
  int pinValue = analogRead(sensorStruct.Pin);
  float pinVolts = 5.0*pinValue/1023.0;
  
  float Pport_abs = (pinVolts - 0.1*sensorStruct.Vsupply) * (sensorStruct.Pmax - sensorStruct.Pmin) / (0.8*sensorStruct.Vsupply) + sensorStruct.Pmin;
  float Pport_gage = psi2kPa(probeatm-Pport_abs)-42;
//  Pport_gage=Pport_gage*0.7366+3.3603;
  return (int) abs(Pport_gage)*10;
} 

float proportionairPressure(int analogPin) {
 double propPress = 0.071*analogRead(analogPin)-proportionairOffset;
 return (float) propPress;
} // END proportionairPressure()

float psi2kPa(float psiVal) {
  return 6.89475729*psiVal;
} // END psi2kPa()

float kpa2psi(float kpaVal) {
  return kpaVal/6.89475729;
} // END psi2kPa()

/************************************************************************/
/************************CALIBRATION_FUNCTION****************************/
/************************************************************************/
void calibrateSystem()
{
/*
PSEUDOCODE:
Arduino sets low-end pressure
Arduino tells PC to enter pressure from both pressure guages
PC enters responds with two values
Arduino sets high-end pressure
Arduino tells PC to enter pressure from both pressure guages
PC enters responds with two values
Calibration routine complete
*/
}

/************************************************************************/
/**********************INITIALIZATION_FUNCTIONS**************************/
/************************************************************************/

// set all pins as appropriate for arduino at startup
void initPins()
{
  if (debugLevel>DEBUG_SOME) Serial.println(F("DEBUG: subroutine initPins() started"));
  for(int pinNum=PINOFFSET_FIRST; pinNum<=PINOFFSET_LAST; pinNum++) {
    switch (pinNum) {
      case PIN_MUX1_B1:
      case PIN_MUX1_B2:
      case PIN_MUX1_B3:
      case PIN_MUX1_B4:
      case PIN_MUX2_B1:
      case PIN_MUX2_B2:
      case PIN_MUX2_B3:
      case PIN_MUX2_B4:
      case PIN_SOLENOID_POWER:
      case PIN_PUMP_POWER:
      case PIN_MUX_ENABLE:
      case PIN_PROBE_LED_RED:
      case PIN_PROBE_LED_GREEN:
      case PIN_BOX_LED:
      case PIN_MUX_TRIGGER:
      case PIN_PROP_WRITE:
        pinMode(pinNum,OUTPUT);
        digitalWrite(pinNum,LOW);
        break;
      case PIN_PROBE_PRESSURE:
      case PIN_PUMP_PRESSURE:
      case PIN_PROP_READ:
        pinMode(pinNum,INPUT);
        break;
      case PIN_PROBE_SWITCH:
        pinMode(pinNum,INPUT_PULLUP);
        break;
    }
    delay(1);
  }
}

void initBools() {
  bElectrodePairsReceived = false;
  bPreConParamsReceived = false;
  bPreConCompleted = false;
  bTestRunning = false;
  bAborted = false;
  bAtmoCal = false;
}

void initStructures() {
  // Pressure specs are in psi
  // specs from datasheet for honeywell pressure sensors  
  pressureSensor_probe.Vsupply = 5;
  pressureSensor_probe.Pmax = 15;
  pressureSensor_probe.Pmin = 0;
  pressureSensor_probe.Pin = PIN_PROBE_PRESSURE;
  
  // specs from datasheet for honeywell pressure sensors
  pressureSensor_reservoir.Vsupply = 5;
  pressureSensor_reservoir.Pmax = 15;
  pressureSensor_reservoir.Pmin = 0;
  pressureSensor_reservoir.Pin = PIN_PUMP_PRESSURE;
  
  preConParams.onPressure = 15;
  preConParams.numCycles = 0;
  preConParams.onDuration = 1000;
  preConParams.offDuration = 1000;
}

void initAtmosphere() {
  if (debugLevel>DEBUG_SOME) Serial.println(F("DEBUG: subroutine initAtmosphere() started"));
  int reservoirKpa = 99;
  bAtmoCal = false;
  int probeKpa = 99;
  float tempPsi = 0;
  setProp(99); // open prop all the way
  unsigned long localMillis = millis();
  int sameCount = 0;
  
  int prevPress = honeywellPressure(pressureSensor_reservoir);
Serial.print(F("the very first reservior pressure:"));
Serial.println(prevPress);
  while (millis()-localMillis<10000) {
    delay(250);
    reservoirKpa = honeywellPressure(pressureSensor_reservoir);
Serial.print(F("after the delay, the reservior pressure is:"));
Serial.println(reservoirKpa);
    if ((abs(reservoirKpa-prevPress)<=1) && (reservoirKpa<10)) {Serial.print(F("I am out of the loop!!!"));sameCount++;}
    else {
      sameCount=0;
      prevPress=reservoirKpa;
      Serial.println(F("I am still in the loop now"));
    }  
    if (sameCount>3) break;
  }
  
  reservoirKpa = honeywellPressure(pressureSensor_reservoir);
    Serial.print(F("Atomosphere reservoirKpa: "));
  Serial.println(reservoirKpa);
  probeKpa = honeywellPressureforprobe(pressureSensor_probe);
  Serial.print(F("Atomosphere probeKpa: "));
  Serial.println(probeKpa);
  
  if ( (reservoirKpa>15) || (probeKpa>15) ) {
    Serial.println(F("ERROR: Pressure too high for calibration. Probe tip obstructed?"));
  }
  else {
    bAtmoCal = true;
    proportionairOffset=proportionairPressure(PIN_PROP_READ);
          Serial.print(F("PROPORTIONAIR OFFSET IS: "));
  Serial.println(proportionairOffset);
  }
  
  tempPsi = kpa2psi(reservoirKpa)+atmPsi;
  if (tempPsi>12.1 && tempPsi<16.3) {
    atmPsi = tempPsi; // bounded by a reasonable range to check for errors -- 0.5psi above and below the world records
    bAtmoCal &= true;
  } 
  else {
    Serial.print(F("ERROR: atmospheric pressure out of range ("));
    Serial.print(tempPsi);
    Serial.println(F("/12.1-16.3)"));
  }
  delay(500);
  setProp(0);
  Serial.print(F("DEBUG: Atmospheric pressure set at ")); Serial.print(atmPsi); Serial.println(F("psi."));
  digitalWrite(PIN_BOX_LED,HIGH);
}
/************************************************************************/
/************************ERROR_PRINT_FUNCTIONS***************************/
/************************************************************************/

void debugFreeMem() {
  if (debugLevel>DEBUG_SOME) {
    Serial.print(F("DEBUG: Free memory (SRAM): "));
    Serial.println(freeMemory());
  }
}
