﻿using System.IO.Ports;
using System.Management.Instrumentation;
using System.Management;
using System.ComponentModel;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using BHComBus.Util;
using BHComBus.ArduinoComm;
using System.Threading;
using System.Collections.Concurrent;
using System;
using System.Text;


namespace BHComBus.ArduinoComm
{
   
    public delegate void SerialStringListener(IArduinoDeviceInterface device, string data);
    public delegate void SerialPortFoundListener(IArduinoDeviceInterface device, string portName);
    public delegate void SerialPortWriteListener(IArduinoDeviceInterface device, string data);

    public delegate void DebugMessageListener(IArduinoDeviceInterface device, string data);
    public delegate void ErrorMessageListener(IArduinoDeviceInterface device, string data);
    public delegate void MessageListener(IArduinoDeviceInterface device, string data);

    public delegate void ArduinoStatusChangeListener(IArduinoDeviceInterface device, ArduinoStatus oldStatus, ArduinoStatus newStatus);

    public sealed class ArduinoStatus
    {
        public static ArduinoStatus CONNECTED = new ArduinoStatus("CONNECTED");
        public static ArduinoStatus ERROR = new ArduinoStatus("ERROR");
        public static ArduinoStatus BUSY = new ArduinoStatus("BUSY");
        public static ArduinoStatus NOT_CONNECTED = new ArduinoStatus("NOT CONNECTED");

        private string _Name;
        private ArduinoStatus(string name) {
            _Name = name;
        }

        public override string ToString()
        {
            return _Name;
        }
    }


    public interface IArduinoDeviceInterface
    {
        event SerialStringListener OnReceiveSerialData;
        event SerialPortWriteListener OnSerialPortWrite;
        event DebugMessageListener OnReceiveDebugMessage;
        event ErrorMessageListener OnReceiveErrorMessage;
        event MessageListener OnReceiveMessage;
        event ArduinoStatusChangeListener OnArduinoStatusChange;

        SerialPort Port { get; }
        ArduinoStatus Status { get; }

        void sendString(string msg);
        void close();
    }

    public class ArduinoNanoDeviceInterface : IArduinoDeviceInterface 
    {
        public const int DEFAULT_BAUD_RATE = 57600;
       
        public static string getArduinoCOMPortName() {
            foreach (COMPort p in COMUtils.GetComPorts())
            {
                if (p.HardwareName.Contains("VCP"))
                {
                    return p.portNum;
                }
            }
            return null;
        }

        public static IEnumerable<string> getCOMPortNames()
        {
            //return from port in COMUtils.GetComPorts() where port.HardwareName.Contains("VCP") select port.portNum;
            return from port in COMUtils.GetComPorts() select port.portNum;
        }

        public event SerialStringListener OnReceiveSerialData;
        public event SerialPortWriteListener OnSerialPortWrite;
        public event DebugMessageListener OnReceiveDebugMessage;
        public event ErrorMessageListener OnReceiveErrorMessage;
        public event MessageListener OnReceiveMessage;
        public event ArduinoStatusChangeListener OnArduinoStatusChange;

        public static ArduinoNanoDeviceInterface currentDevice;

        public static void SetDevice(string portname)
        {
            if (currentDevice != null)
            {
                if (portname == currentDevice.port.PortName) return;
                currentDevice.close();
            }
            currentDevice = new ArduinoNanoDeviceInterface(portname, DEFAULT_BAUD_RATE);
            Global_Varible.StopFlagBoolean = false;
            //currentDevice.port.Open();

        }

        private ArduinoStatus _status;
        public ArduinoStatus Status
        {
            get
            {
                return _status;
            }
            set
            {
                if (value != _status)
                {
                    var oldStatus = _status;
                    _status = value;
                    var oASC = OnArduinoStatusChange;
                    if (oASC != null)
                        oASC(this, oldStatus, _status);
                }
            }
        }

        private SerialPort port;
        public SerialPort Port
        {
            get
            {
                return port;
            }

            private set { port = value; }
        }

        private ConcurrentQueue<string> portReadLines;
        private ConcurrentQueue<string> portWriteLines;
        private Thread ioThread;
        private bool closeThread = false;

        public ArduinoNanoDeviceInterface(string ComPortName, int baudRate )
        {
            port = new SerialPort(ComPortName);
            port.Encoding = System.Text.Encoding.ASCII;
            port.BaudRate = baudRate;
            portReadLines = new ConcurrentQueue<string>();
            portWriteLines = new ConcurrentQueue<string>();

            Status = ArduinoStatus.NOT_CONNECTED;

            ioThread = new Thread(port_IOLoop);
            ioThread.IsBackground = true;
            ioThread.Priority = ThreadPriority.BelowNormal;
            ioThread.Start();
        }

        public void close() {
            closeThread = true;
        }

        void port_ErrorReceived(object sender, SerialErrorReceivedEventArgs e)
        {
            var oREM = OnReceiveErrorMessage;
            if (oREM != null)
            {
                oREM(this, "SerialPort Error Received: " + e.EventType.ToString());
            }
        }

        void port_IOLoop()
        {
            
            port.ErrorReceived += port_ErrorReceived;
            StringBuilder csv = new StringBuilder();
            string currentProbePressure = string.Empty;
            string laterProbePressure = string.Empty;
            string laterPropPressure = string.Empty;
            while (true)
            {
                if (closeThread)
                {
                    if (port.IsOpen)
                        port.Close();
                    ioThread = null;
                    return;
                }
                
                try
                {


                    if (!port.IsOpen)
                    {
                        if (!SerialPort.GetPortNames().Contains(port.PortName))
                        {
                            // The Arduino is no longer plugged in?
                            Status = ArduinoStatus.NOT_CONNECTED;
                            Thread.Sleep(200);
                            continue;
                        }
                        port.Open();
                        if(port.IsOpen)
                            Status = ArduinoStatus.CONNECTED;
                    }
                }
                catch (System.Exception ex)
                {
                    // Couldn't open the port for whatever reason
                    Status = ArduinoStatus.NOT_CONNECTED;
                    System.Threading.Thread.Sleep(200);
                    continue;
                }
                var readData = false;
                var wroteData = false;

                try
                {
                    if (port.IsOpen && port.BytesToRead > 0)
                    {
                        portReadLines.Enqueue(port.ReadLine());
                        string data;
                        portReadLines.TryDequeue(out data);
                        if (data != null)
                        {
                            readData = true;
                            var oRSD = OnReceiveSerialData;
                            if (oRSD != null)
                                oRSD(this, data);
                            if (data.StartsWith("ERROR:"))
                            {
                                var oREM = OnReceiveErrorMessage;
                                if (oREM != null)
                                    oREM(this, data);
                            }
                            else if (data.StartsWith("DEBUG:"))
                            {
                                var oRDM = OnReceiveDebugMessage;
                                if (oRDM != null)
                                    oRDM(this, data);
                             //   string aa = data;
                            //   System.IO.File.WriteAllText(@"D:\MyData.txt", aa);
                                //new System.IO.StreamWriter("D:\\MyTest.txt");
                              //  var aa = data;
                              //  file.WriteLine(aa);
                               // System.IO.File.AppendText.WriteLine(@"D:\MyTest.txt", aa);
                            }
                            else if (data.StartsWith("currentprobepressure:"))
                            {
                                currentProbePressure = string.Empty;
                             //   string aa = string.Empty;
                                for (int i = 0; i < data.Length; i++)
                                {
                                    if (char.IsDigit(data[i]))
                                        currentProbePressure += data[i];                                
                                }
                           //     aa =aa+ "\r\n";
                            //    System.IO.File.AppendAllText(@"D:\currentprobepressure.txt", currentProbePressure);
                            }
                            else if (data.StartsWith("laterprobepressure:"))
                            {
                                laterProbePressure = string.Empty;
                                string aa = string.Empty;
                                for (int i = 0; i < data.Length; i++)
                                {
                                    if (char.IsDigit(data[i]))
                                    {
                                        laterProbePressure += data[i];
                                        aa += data[i];
                                    }
                                }
                                aa = aa + "\r\n";
                                System.IO.File.AppendAllText(@"D:\laterprobepressure.txt", aa);

                            }


                            else if (data.StartsWith("propsetpressure:"))
                            {
                                string aa = string.Empty;
                                for (int i = 0; i < data.Length; i++)
                                {
                                    if (char.IsDigit(data[i]))
                                        aa += data[i];
                                }
                                aa = aa + "\r\n";
                                System.IO.File.AppendAllText(@"D:\propsetpressure.txt", aa);
                            }
                            else if (data.StartsWith("desiredprobepressure:"))
                            {
                                string aa = string.Empty;
                                for (int i = 0; i < data.Length; i++)
                                {
                                    if (char.IsDigit(data[i]))
                                        aa += data[i];
                                }
                                aa = aa + "\r\n";
                                System.IO.File.AppendAllText(@"D:\desiredprobepressure.txt", aa);
                            }
                            else if (data.StartsWith("desiredproppressure:"))
                            {
                                string aa = string.Empty;
                                for (int i = 0; i < data.Length; i++)
                                {
                                    if (char.IsDigit(data[i]))
                                        aa += data[i];
                                }
                                aa = aa + "\r\n";
                                System.IO.File.AppendAllText(@"D:\desiredproppressure.txt", aa);
                            }

                            else if (data.StartsWith("laterproppressure:"))
                            {
                           //     string aa = string.Empty;
                                laterPropPressure = string.Empty;
                                for (int i = 0; i < data.Length; i++)
                                {
                                    if (char.IsDigit(data[i])||data[i]=='.')
                                        laterPropPressure += data[i];
                                }
                       //         aa = aa + "\r\n";
                              //  System.IO.File.AppendAllText(@"D:\laterproppressure.txt", aa);
                                var newLine = string.Format("{0}, {1}, {2},{3}", currentProbePressure, laterProbePressure, laterPropPressure, Environment.NewLine);
                                csv.Append(newLine);
                            }

                            else 
                            {
                                var oRM = OnReceiveMessage;
                                if (oRM != null)
                                    oRM(this, data);
                            }
                            // Adding new rows to csv StringBuilder
                            //var newLine = string.Format("{0}, {1}, {2}", currentProbePressure, laterProbePressure, Environment.NewLine);
                            //csv.Append(newLine);
                        }
                    }

                    if (portWriteLines.Count > 0 && port.IsOpen)
                    {
                        string msg = null;
                        if (portWriteLines.TryDequeue(out msg))
                        {
                            wroteData = true;
                            msg = msg.Trim();
                            var msgBytes = System.Text.Encoding.ASCII.GetBytes(msg);
                            port.Write(msgBytes, 0, msgBytes.Length);
                            var oSPW = OnSerialPortWrite;
                            if (oSPW != null)
                                oSPW(this, msg);
                        }
                    }
                }

                catch (System.Exception ex)
                {
                    Debug.WriteLine("Encountered error during Arduino IO operatinos");
                }
                if (!readData && !wroteData)
                { // Writing StringBuilder to .csv file
                    System.IO.File.WriteAllText(@"D:\probPressures.csv", csv.ToString());
                    Thread.Sleep(50);
                }


               

            }

        }

        public void sendString(string msg) {
            //passing write through a concurrent queue to eliminate issues when writing from multiple
            //threads
            portWriteLines.Enqueue(msg);
        }

    }

    public class Manager
    {
        
    }
}
