﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using BHComBus.ArduinoComm;
using BHComBus.Data;
using BHComBus.Util;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Threading;
using BHComBus.FGU;
using System.IO.Ports;



namespace BHComBus
{
    public class ProtocolManager : INotifyPropertyChanged
    {

        #region INotifyPropertyChanged Implementation

        public event PropertyChangedEventHandler PropertyChanged;

        public void InvokePropertyChanged(PropertyChangedEventArgs e)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null) handler(this, e);
        }

        // Convenience method
        public void InvokePropertyChanged(string property)
        {
            InvokePropertyChanged(new PropertyChangedEventArgs(property));
        }
        #endregion

        private IArduinoDeviceInterface _arduino;
        public IArduinoDeviceInterface Arduino
        {
            get { return _arduino; }
            set {
                if (_arduino != null)
                {
                    if (_arduino == value) return;
                    _arduino.OnReceiveMessage -= _arduino_OnReceiveMessage;
                }
                
                _arduino = value;
                _arduino.OnReceiveMessage += _arduino_OnReceiveMessage;
            }
        }

        public ProtocolManager(IArduinoDeviceInterface device)
        {
            _arduino = device;

            _arduino.OnReceiveMessage += _arduino_OnReceiveMessage;
            
        }

        private ProtocolState _state;
        public ProtocolState State
        {
            get { return _state;  }
            private set { 
                _state = value;
                InvokePropertyChanged(new PropertyChangedEventArgs("State"));
            }
        }

        private ExperimentRecipe _recipe;
        public ExperimentRecipe Recipe
        {
            get { return _recipe; }
            set {
                _recipe = value;
                InvokePropertyChanged("Recipe");
                resetState();
            }
        }

        private int _curElectrodeIdx = 0;
        private int _curSamplingParamIdx = 0;

        public ProtocolManager()
        {
            State = ProtocolState.PRE_INIT;
        }

        private void resetState()
        {

            _curElectrodeIdx = 0;
            State = ProtocolState.PRE_INIT;
        }

        /********************  Program API *****************/
        public void run()
        {
            sendMessage(PCMessage.SET_VERBOSITY(0));
            Thread.Sleep(150);
            sendPrecondition();
        }

        public void sendPrecondition()
        {
            var pre = Recipe.preConditioningConfiguration;
            sendMessage(PCMessage.SET_PRECONDITION(pre.Pressure_kPa, pre.CycleCount, pre.TimeOn_ms, pre.TimeOff_ms));
        }

        public void sendElectrodes()
        {
            sendMessage(PCMessage.SET_ELECTRODES(Recipe.ElectrodePairsList.Count));
        }
   
        /******************** End Program API **************/

        void _arduino_OnReceiveMessage(IArduinoDeviceInterface device, string data)
        {
            /*
             * 
             */
            ArduinoMessage msg = ArduinoMessage.Parse(data);

            switch (State)
            {
                case ProtocolState.PRE_INIT:
                    PRE_INIT(msg); return;
                default:
                    return;
            }
            
        }

        //Alex + Zhenchen
        public static void p_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            Console.WriteLine(
                (sender as SerialPort).ReadExisting());
        }

        public String portMsg;
        
        private int testCount = 0;
        private ExperimentState state;
        private bool _electrodesConfigured;
        private void PRE_INIT(ArduinoMessage msg)
        {
            MuxHoldSettings nextSet;
            switch (msg.Type)
            {
                case MessageType.ARDUINO_NOTIFY_PRECON_COMPLETE:
                    if (!COMUtils.FGPort.IsOpen)
                    {
                        COMUtils.FGPort.Open();   
                    }
                    COMUtils.FGPort.Write("1" + "\r\n");        
                    Thread.Sleep(30);

                    state = new ExperimentState(Recipe);
                    nextSet = state.getNextMuxHoldSetting();
                    sendMessage(PCMessage.MUX_HOLD_HECTO_PASCALS(nextSet.hold_us, nextSet.pressure_hPa, nextSet.propDelay_ms, nextSet.electrodePair.E1, nextSet.electrodePair.E2));
                    testCount = 0;
                    testCount++;
                    return;
                case MessageType.ARDUINO_NOTIFY_MUX_HOLD_COMPLETE:
                    
                    nextSet = state.getNextMuxHoldSetting();
                    if (nextSet == null || Global_Varible.StopFlagBoolean==true)
                    {
                        Thread.Sleep(50);
                        sendMessage(PCMessage.RESET());
                        COMUtils.FGPort.Close();
                        Global_Varible.StopFlagBoolean = false;
                        return;
                    }     
                    else 
                    {   
                        
                        //8 is the number of electrode pairs tested for each frequency
                        if (testCount >= 8)
                        {
                            //Switching frequency for Mingyuan's function generator - Alex + Zhenchen
                            portMsg = Convert.ToString((1.0f * (100 * 1000000) / nextSet.hold_us)/1000);
                            
                            if (portMsg.Length != 0)
                            
                            {
                               
                                COMUtils.FGPort.Write(portMsg + "\r\n");
                                Thread.Sleep(30);
                                
                            }
                            //Reset testCount
                            testCount = 0;
                            
                        }
                                                
                        sendMessage(PCMessage.MUX_HOLD_HECTO_PASCALS(nextSet.hold_us, nextSet.pressure_hPa, nextSet.propDelay_ms, nextSet.electrodePair.E1, nextSet.electrodePair.E2));

                        testCount++;
                    }
                    return;
                case MessageType.ARDUINO_CONFIRM_BIOIMPEDANCE_PARAMS_RECEIVED:
                                         
                    sendMessage(PCMessage.RUN_READY());
                    return;
                case MessageType.ARDUINO_NOTIFY_STOP:
                    Global_Varible.StopFlagBoolean = true;
                    return;
                    
                case MessageType.ARDUINO_CONFIRM_PRECON_CONFIGURED:
                    if (Recipe.hasElectrodePairs())
                        sendElectrodes();
                    return;
                case MessageType.ARDUINO_CONFIRM_ELECTRODE_PAIR_RECEIVED:
                case MessageType.ARDUINO_REQUEST_ELECTRODE_PAIR:
                    if (Recipe == null || !Recipe.hasElectrodePairs())
                    {
                        sendMessage(PCMessage.ABORT());
                    }
                    else if (_curElectrodeIdx >= Recipe.ElectrodePairsList.Count)
                    {
                        sendMessage(PCMessage.ELECTRODES_COMPLETE());
                        _curElectrodeIdx = 0;
                    }
                    else
                    {
                        var ep = Recipe.ElectrodePairsList[_curElectrodeIdx];
                        sendMessage(PCMessage.SET_ELECTRODE_PAIR(_curElectrodeIdx, ep.E1, ep.E2));
                        _curElectrodeIdx++;
                    }

                    return;

                case MessageType.ARDUINO_NOTIFY_ELECTRODE_PAIRS_RECEIVED:
                    _electrodesConfigured = true;
                    return;

                case MessageType.ARDUINO_ERROR:
                    State = ProtocolState.PRE_INIT;
                    sendMessage(PCMessage.ABORT());
                    return;

                case MessageType.ARDUINO_NOTIFY_ABORT:
                    //Nothing to do here
                    return;
                
                default:
                    return;
            }
        }


        private void sendMessage(PCMessage msg)
        {
            Arduino.sendString(msg);
        }
        
    }

    public enum ProtocolState
    {
        PRE_INIT,
        CONFIG,
        BUSY,
        PRECONDITIONING,
        TESTING,
        ERROR
    }


  
}
      