﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.ComponentModel.DataAnnotations;
using FileHelpers;
using BHComBus.FGU;
//using System.Linq;

namespace BHComBus.Data
{
    public static class BH_CONSTANTS
    {
        public const int MIN_PRESSURE = 5;
        public const int MAX_PRESSURE = 45;

        public const int MIN_FREQ = 100;
        public const int MAX_FREQ = 1000000;

        public const int MIN_SAMPLE_TIME = 10;
        public const int MAX_SAMPLE_TIME = 2147483647;

        public const int MIN_PRECON_HOLD_TIME = 0;
        public const int MAX_PRECON_HOLD_TIME = 6000;

        public const int MIN_PRECON_CYCLE_COUNT = 0;
        public const int MAX_PRECON_CYCLE_COUNT = 6;
    }

    [DelimitedRecord(",")]
    public class ElectrodePair
    {
        [Range(0, 7)]
        public int E1;

        [Range(0, 7)]
        public int E2;

        public ElectrodePair()
        {
            E1 = 0;
            E2 = 0;
        }
        public ElectrodePair(int e1, int e2)
        {
            E1 = e1;
            E2 = e2;
        }
    }

    [DelimitedRecord(",")]
    public class FunctionGeneratorSettings
    {
        public int Frequency_Hz;
        public int Amplitude_mV;
        public Function Function;

        public FunctionGeneratorSettings() { }
        public FunctionGeneratorSettings(Function func, int freq_hz, int ampl_mV)
        {
            Function = func;
            Frequency_Hz = freq_hz;
            Amplitude_mV = ampl_mV;
        }
    }

    [DelimitedRecord(",")]
    public class SamplingParameters
    {
        [Range(100, 1000000, ErrorMessage = "Valid frequencies are between 100hz and 1Mhz")]
        public int Frequency_Hz;

        //Pressure to hold during bioimpedance testing at this frequency (in kPa)
        public float Pressure_kPa;

        //Time, in microseconds, to test each electrode pair (i.e., roughly the time the Arduino should hold the MUX enable line high)
        //private long _pairTime_us;
        public int PairTime_us
        {
            get;
            set;
        }

        public SamplingParameters()
        {

        }

        public SamplingParameters(int freq_hz, float pressure_kPa, int pairTime_us)
        {
            Frequency_Hz = freq_hz;
            Pressure_kPa = pressure_kPa;
            PairTime_us = pairTime_us;
        }
        

        public static long holdTimeRequired_us(long freq_hz, int periodCount)
        {
            return (periodCount * 1000000) / freq_hz;
        }

        public static long sampleIntervalRequired_ns(long freq_hz, int samplesPerPeriod)
        {
            return 1000000000 / (freq_hz * samplesPerPeriod);
        }

        public static FileHelperAsyncEngine<SamplingParameters> DefaultEngine = new FileHelperAsyncEngine<SamplingParameters>();
    }

    [DelimitedRecord(",")]
    public class BioImpedanceDataPoint
    {
        public int SetNumber;
        public uint Frequency_Hz;
        public float Pressure_kPa;
        public ulong HoldTime_us;
        public uint SampleInterval_ns;
        public int TriggerTimeOffset_ms;
        public int PairNumber;
        public byte E1;
        public byte E2;
        public short ChanA_raw;
        public short ChanB_raw;
        public double ChanA_V;
        public double ChanB_V;

        public static FileHelperAsyncEngine<BioImpedanceDataPoint> DefaultEngine = new FileHelperAsyncEngine<BioImpedanceDataPoint>();
    }

    public class ConditioningConfig
    {
        public float Pressure_kPa;
        public int TimeOn_ms;
        public int TimeOff_ms;
        public int CycleCount;
    }

    public class WaveFormData
    {
        public uint Range_mV;
        public short[] ChanA;
        public short[] ChanB;
    }

    public class MuxHoldSettings
    {
        public MuxHoldSettings(ElectrodePair e, int h, int del, int pres)
        {
            electrodePair = e;
            hold_us = h;
            propDelay_ms = del;
            pressure_hPa = pres;
        }
        public ElectrodePair electrodePair;
        public int hold_us;
        public int propDelay_ms;
        public int pressure_hPa;
    }

    public class ExperimentState
    {

        public ExperimentState(ExperimentRecipe rec)
        {
            //Initialize indices to 0
            recipe = rec;
            currentPressureIdx = 0;
            currentElectrodeIdx = 0;
            currentFrequencyIdx = 0;
        }

        //Declare variables
        public ExperimentRecipe recipe;
        public int currentPressureIdx;
        public int currentElectrodeIdx;
        public int currentFrequencyIdx;

        public MuxHoldSettings getNextMuxHoldSetting()
        {
           
            if(recipe != null 
               && recipe.ElectrodePairsList != null
               && recipe.PressuresHpaList != null
               && currentPressureIdx < recipe.PressuresHpaList.Count) {
                   var delayTime = (currentFrequencyIdx == 0 && currentElectrodeIdx == 0) ? 50 : 0;
                var set = new MuxHoldSettings(recipe.ElectrodePairsList[currentElectrodeIdx], (int) SamplingParameters.holdTimeRequired_us(recipe.FrequencyHzList[currentFrequencyIdx],100), delayTime, recipe.PressuresHpaList[currentPressureIdx]);
                currentElectrodeIdx++;
                //If electrode number is greater than total, go to next frequency and reset electrode number
                if (currentElectrodeIdx >= recipe.ElectrodePairsList.Count)
                {
                    currentElectrodeIdx = 0;
                    currentFrequencyIdx++;
                    //currentPressureIdx++;
                }

                //If frequency number list has been iterated, go to next pressure and reset frequency list
                if (currentFrequencyIdx >= recipe.FrequencyHzList.Count)
                {
                    currentFrequencyIdx = 0;
                    currentPressureIdx++;
                }
                return set;
            }
            return null;
        }
    }

    public class ExperimentRecipe
    {
        public ConditioningConfig preConditioningConfiguration;
        public List<SamplingParameters> SamplingParametersList;
        public List<ElectrodePair> ElectrodePairsList;
        public List<int> PressuresHpaList;
        public List<int> PressureControlHpaList;
        //public int FrequencyHz;
        public List<int> FrequencyHzList;

        public bool hasElectrodePairs()
        {
            return ElectrodePairsList != null && ElectrodePairsList.Count > 0;
        }

        public bool canRunBioimpedance()
        {
            return SamplingParametersList != null && SamplingParametersList.Count > 0
                && ElectrodePairsList != null & ElectrodePairsList.Count > 0;
        }

        public bool canRunPreconditioning()
        {

            return preConditioningConfiguration != null
                && preConditioningConfiguration.Pressure_kPa >= BH_CONSTANTS.MIN_PRESSURE
                && preConditioningConfiguration.Pressure_kPa <= BH_CONSTANTS.MAX_PRESSURE
                && preConditioningConfiguration.TimeOff_ms >= BH_CONSTANTS.MIN_PRECON_HOLD_TIME 
                && preConditioningConfiguration.TimeOff_ms <= BH_CONSTANTS.MAX_PRECON_HOLD_TIME
                && preConditioningConfiguration.TimeOn_ms > BH_CONSTANTS.MIN_PRECON_HOLD_TIME
                && preConditioningConfiguration.TimeOn_ms <= BH_CONSTANTS.MAX_PRECON_HOLD_TIME
                && preConditioningConfiguration.CycleCount >= BH_CONSTANTS.MIN_PRECON_CYCLE_COUNT
                && preConditioningConfiguration.CycleCount <= BH_CONSTANTS.MAX_PRECON_CYCLE_COUNT; //0 is allowed for CycleCount

        }

        public bool canRunFull()
        {
            return canRunBioimpedance() && canRunPreconditioning();
        }
    }
}
